import app from ":virtual/vitebook/app";
import {
  DATA_ASSET_BASE_PATH,
  isBoolean,
  isLoadedMarkdownPage,
  isString,
  resolveDataAssetID
} from "../shared";
import { createRouter } from "./router/create-router";
import { createMemoryHistory } from "./router/history/memory";
async function initRouter({
  $route,
  $navigation,
  $pages,
  $layouts,
  $currentPage,
  serverContext
}) {
  const router = createRouter({
    baseUrl: app.baseUrl,
    trailingSlash: !import.meta.env.SSR ? window["__VBK_TRAILING_SLASH__"] : true,
    history: !import.meta.env.SSR ? window.history : createMemoryHistory(),
    $route,
    $navigation,
    serverContext
  });
  if (!import.meta.env.SSR && import.meta.env.PROD) {
    const redirects = window["__VBK_REDIRECTS_MAP__"] ?? {};
    for (const from of Object.keys(redirects)) {
      const to = redirects[from];
      router.addRedirect(from, to);
    }
  }
  await Promise.all(app.configs.map((config) => config?.({ router })));
  addRoutes(router, $pages.get(), $layouts.get(), $currentPage);
  if (import.meta.hot) {
    handlePageHMR($currentPage);
    handlePagesHMR(router, $pages, $layouts, $currentPage);
  }
  return router;
}
let routes = [];
function addRoutes(router, pages, layouts, $currentPage) {
  for (const page of pages) {
    const route = {
      ...page.route,
      loader: async ({ redirect }) => {
        const pageOrRedirect = await loadPage(
          router,
          page,
          layouts,
          $currentPage
        );
        return isString(pageOrRedirect) ? redirect(pageOrRedirect) : pageOrRedirect;
      },
      prefetch: async ({ url, redirect }) => {
        const redirectTo = await loadPage(router, page, layouts, $currentPage, {
          prefetch: url
        });
        return isString(redirectTo) ? redirect(redirectTo) : void 0;
      }
    };
    router.addRoute(route);
    routes.push(route);
  }
}
function handlePageHMR($currentPage) {
  if (import.meta.hot) {
    import.meta.hot.on("vitebook::md_meta", ({ filePath, meta }) => {
      const page = $currentPage.get();
      if (isLoadedMarkdownPage(page) && filePath.endsWith(page.rootPath)) {
        $currentPage.set({ ...page, meta });
      }
    });
    $currentPage.subscribe(($page) => {
      import.meta.hot.send("vitebook::page_change", {
        rootPath: $page?.rootPath ?? ""
      });
    });
  }
}
function handlePagesHMR(router, $pages, $layouts, $currentPage) {
  if (import.meta.hot) {
    const updatePages = (pages) => {
      routes.forEach((route) => router.removeRoute(route));
      routes = [];
      addRoutes(router, pages, $layouts.get(), $currentPage);
    };
    let pagesUpdateTimer;
    const delayedPagesUpdate = (pages) => {
      clearTimeout(pagesUpdateTimer);
      pagesUpdateTimer = setTimeout(() => updatePages(pages), 100);
    };
    $pages.subscribe(($pages2) => delayedPagesUpdate($pages2));
    $layouts.subscribe(() => delayedPagesUpdate($pages.get()));
  }
}
async function loadPage(router, clientPage, clientLayouts, $currentPage, { prefetch = false } = {}) {
  const prefetchURL = isBoolean(prefetch) ? void 0 : prefetch;
  let pageModule;
  let pageData;
  let layouts;
  if (import.meta.env.DEV) {
    pageModule = await clientPage.loader();
    pageData = await loadData(router, pageModule, void 0, prefetchURL);
    if (import.meta.env.DEV && isString(pageData.__redirect__)) {
      return pageData.__redirect__;
    }
    layouts = await loadPageLayouts(
      router,
      clientPage,
      clientLayouts,
      prefetchURL
    );
  } else {
    [pageModule, layouts] = await Promise.all([
      (async () => {
        const mod = await clientPage.loader();
        pageData = await loadData(router, mod, void 0, prefetchURL);
        return mod;
      })(),
      loadPageLayouts(router, clientPage, clientLayouts, prefetchURL)
    ]);
  }
  const loadedPage = {
    ...clientPage,
    $$loaded: true,
    module: pageModule,
    layouts,
    data: pageData,
    get default() {
      return pageModule.default;
    },
    get meta() {
      return isLoadedMarkdownPage(this) ? pageModule.meta : void 0;
    }
  };
  if (!prefetch) {
    $currentPage.set(loadedPage);
  }
  return loadedPage;
}
function loadPageLayouts(router, page, layouts, prefetchURL) {
  return Promise.all(
    page.layouts.map(async (index) => {
      const layout = layouts[index];
      const mod = await layout.loader();
      return {
        $$loaded: true,
        ...layout,
        module: mod,
        data: await loadData(router, mod, index, prefetchURL),
        get default() {
          return mod.default;
        }
      };
    })
  );
}
async function loadData(router, module, layoutIndex, prefetchURL) {
  if (!module.loader)
    return {};
  let pathname = prefetchURL?.pathname ?? router.navigation.get().to.pathname;
  if (!pathname.endsWith("/"))
    pathname += "/";
  const id = resolveDataAssetID(decodeURI(pathname), layoutIndex);
  const hashId = import.meta.env.PROD && !import.meta.env.SSR ? window["__VBK_DATA_HASH_MAP__"][await hashDataId(id)] : id;
  if (!hashId)
    return {};
  if (import.meta.env.SSR && router.serverContext) {
    return router.serverContext.data.get(hashId) ?? {};
  }
  try {
    const json = !router.started ? getDataFromScript(hashId) : await (await fetch(`${DATA_ASSET_BASE_PATH}/${hashId}.json`)).json();
    return json;
  } catch (e) {
    console.error(e);
  }
  return {};
}
function getDataFromScript(id) {
  const dataScript = document.getElementById("__VBK_DATA__");
  try {
    const content = JSON.parse(dataScript.textContent);
    return content[id] ?? {};
  } catch (e) {
  }
  return {};
}
async function hashDataId(id) {
  const encodedText = new TextEncoder().encode(id);
  const hashBuffer = await crypto.subtle.digest("SHA-1", encodedText);
  const hashArray = Array.from(new Uint8Array(hashBuffer));
  return hashArray.map((b) => b.toString(16).padStart(2, "0")).join("").substring(0, 8);
}
export {
  initRouter,
  loadData,
  loadPage,
  loadPageLayouts
};
