import { isFunction, isString, normalizeURL, slash } from "../../shared";
import { createSimpleRoutesComparator } from "./comparators/comparator";
import { listen } from "./listener";
import {
  createSimpleScrollDelegate
} from "./scroll-delegate";
function createRouter({
  history,
  baseUrl = "/",
  trailingSlash = true,
  $navigation,
  $route,
  initialRoutes = [],
  serverContext
}) {
  let _url, _started = false, _disabled = false, _historyIndex = Date.now(), _routes = [], _currentRoute = null, _scrollDelegate, _comparator = createSimpleRoutesComparator();
  const _historyKey = "vitebook::index", _redirects = /* @__PURE__ */ new Map(), _beforeNavigate = [], _afterNavigate = [];
  if (import.meta.env.SSR) {
    _url = new URL("http://ssr");
  } else {
    _url = new URL(location.href);
    document.body.setAttribute("tabindex", "-1");
  }
  const buildURL = (pathnameOrURL) => {
    const url = !isString(pathnameOrURL) ? new URL(pathnameOrURL) : new URL(
      pathnameOrURL,
      pathnameOrURL.startsWith("#") ? /(.*?)(#|$)/.exec(location.href)[1] : getBaseUri(baseUrl)
    );
    return normalizeURL(url, trailingSlash);
  };
  const owns = (url) => {
    if (import.meta.env.SSR)
      return true;
    return url.origin === location.origin && url.pathname.startsWith(baseUrl);
  };
  const hasRoute = (pathnameOrRoute) => {
    const url = buildURL(
      !isString(pathnameOrRoute) ? pathnameOrRoute.pathname : pathnameOrRoute
    );
    return _comparator.match(url, _routes) !== null;
  };
  const findRoute = (pathnameOrURL) => {
    const url = isString(pathnameOrURL) ? buildURL(pathnameOrURL) : pathnameOrURL;
    if (owns(url)) {
      const route = _comparator.match(url, _routes);
      if (!route)
        return null;
      const pathname = decodeURI(slash(url.pathname.slice(baseUrl.length)));
      const query = new URLSearchParams(url.search);
      const id = `${pathname}?${query}`;
      return { ...route, id, url };
    }
    return null;
  };
  const goLocation = async (url) => {
    location.href = url.href;
    return new Promise(() => {
    });
  };
  function _redirect(url, handle) {
    const href = normalizeURL(url, trailingSlash).href;
    if (!_redirects.has(href))
      return null;
    const redirectHref = _redirects.get(href);
    const redirectURL = new URL(redirectHref);
    if (owns(redirectURL)) {
      return handle(redirectURL);
    } else {
      return goLocation(redirectURL);
    }
  }
  const addRedirect = (from, to) => {
    const fromURL = normalizeURL(
      isString(from) ? buildURL(from) : from,
      trailingSlash
    );
    const toURL = normalizeURL(isString(to) ? buildURL(to) : to, trailingSlash);
    _redirects.set(fromURL.href, toURL.href);
  };
  function _buildRedirect(from, handleRedirect) {
    const fromURL = buildURL(from);
    return async (pathOrURL) => {
      const redirectURL = buildURL(pathOrURL);
      addRedirect(fromURL, redirectURL);
      await handleRedirect(redirectURL);
    };
  }
  const prefetch = async (pathnameOrURL) => {
    const url = buildURL(pathnameOrURL);
    let redirecting = _redirect(url, (url2) => prefetch(url2));
    if (redirecting)
      return redirecting;
    const route = findRoute(url);
    if (!route) {
      throw new Error(
        "Attempted to prefetch a URL that does not belong to this app"
      );
    }
    const redirect = _buildRedirect(pathnameOrURL, (redirectURL) => {
      redirecting = prefetch(redirectURL);
    });
    await route.prefetch?.({
      route,
      url,
      redirect
    });
    if (redirecting)
      return redirecting;
  };
  const changeHistoryState = (url, state, replace) => {
    const change = replace ? 0 : 1;
    state[_historyKey] = _historyIndex += change;
    history[replace ? "replaceState" : "pushState"](state, "", url);
  };
  if (!import.meta.env.SSR) {
    _historyIndex = history.state?.[_historyKey];
    if (!_historyIndex) {
      _historyIndex = Date.now();
      history.replaceState(
        {
          ...history.state,
          [_historyKey]: _historyIndex
        },
        "",
        location.href
      );
    }
  }
  async function navigate(url, {
    scroll,
    accepted,
    blocked,
    state = {},
    keepfocus = false,
    replace = false
  }) {
    let cancelled = false;
    const cancel = () => {
      $navigation.set(null);
      if (!cancelled)
        blocked?.();
      cancelled = true;
    };
    const redirectNavigation = (url2) => navigate(url2, {
      keepfocus,
      replace,
      state,
      accepted,
      blocked: cancel
    });
    let redirecting = _redirect(url, redirectNavigation);
    if (redirecting)
      return redirecting;
    if (!owns(url)) {
      cancel();
      return;
    }
    const route = findRoute(url);
    if (!route) {
      cancel();
      throw new Error(
        "Attempted to navigate to a URL that does not belong to this app"
      );
    }
    const redirect = _buildRedirect(url, (redirectURL) => {
      redirecting = redirectNavigation(redirectURL);
    });
    for (const hook of _beforeNavigate) {
      hook({
        from: _currentRoute,
        to: route,
        params: route.params,
        cancel,
        redirect
      });
    }
    if (cancelled)
      return;
    if (redirecting)
      return redirecting;
    _scrollDelegate.savePosition?.();
    accepted?.();
    $navigation.set({ from: url, to: url });
    const page = await route.loader({ route, redirect });
    if (redirecting)
      return redirecting;
    if (!page) {
      cancel();
      return;
    }
    const fromRoute = _currentRoute;
    const toRoute = { ...route, page };
    _currentRoute = toRoute;
    $route?.set(toRoute);
    if (!import.meta.env.SSR) {
      await new Promise((res) => window.requestAnimationFrame(res));
    }
    changeHistoryState(url, state, replace);
    if (!import.meta.env.SSR) {
      if (!keepfocus) {
        const root = document.body;
        const tabindex = root.getAttribute("tabindex");
        getSelection()?.removeAllRanges();
        root.tabIndex = -1;
        root.focus({ preventScroll: true });
        if (tabindex !== null) {
          root.setAttribute("tabindex", tabindex);
        } else {
          root.removeAttribute("tabindex");
        }
      }
      await new Promise((res) => window.requestAnimationFrame(res));
      await _scrollDelegate.scroll?.({
        from: fromRoute,
        to: toRoute,
        target: scroll,
        hash: url.hash
      });
    }
    await Promise.all(
      _afterNavigate.map(
        (hook) => hook({
          from: fromRoute,
          to: toRoute,
          params: route.params
        })
      )
    );
    _url = url;
    $navigation.set(null);
  }
  const hashChanged = (hash) => {
    _url.hash = hash;
    if (_currentRoute) {
      $route?.set({
        ..._currentRoute,
        url: _url
      });
    }
  };
  const router = {
    baseUrl,
    trailingSlash,
    serverContext,
    get url() {
      return _url;
    },
    get started() {
      return _started;
    },
    get currentRoute() {
      return _currentRoute;
    },
    get _historyKey() {
      return _historyKey;
    },
    get _historyIndex() {
      return _historyIndex;
    },
    set _historyIndex(index) {
      _historyIndex = index;
    },
    get scroll() {
      return _scrollDelegate;
    },
    get navigation() {
      return $navigation;
    },
    get navigating() {
      return !!$navigation.get();
    },
    get disabled() {
      return _disabled;
    },
    set disabled(disabled) {
      _disabled = disabled;
    },
    buildURL,
    owns,
    hasRoute,
    findRoute,
    goLocation,
    prefetch,
    addRedirect,
    hashChanged,
    _navigate: navigate,
    beforeNavigate(hook) {
      _beforeNavigate.push(hook);
    },
    afterNavigate(hook) {
      _afterNavigate.push(hook);
    },
    start() {
      if (!_started) {
        listen(router, history);
        _started = true;
      }
    },
    getRoute(pathname) {
      const url = buildURL(pathname);
      return _comparator.match(url, _routes);
    },
    addRoute(route) {
      if (!hasRoute(route)) {
        _routes.push({
          ...route,
          score: route.score ?? _comparator.score(route)
        });
        _routes = _comparator.sort(_routes);
      }
    },
    removeRoute(route) {
      _routes = _routes.filter((r) => r !== route);
    },
    back() {
      history.back();
    },
    async go(path, { scroll, replace = false, keepfocus = false, state = {} } = {}) {
      if (isString(path) && path.startsWith("#")) {
        const hash = path;
        hashChanged(hash);
        changeHistoryState(_url, state, replace);
        await _scrollDelegate.scroll?.({ target: scroll, hash });
        _scrollDelegate.savePosition?.();
        return;
      }
      const url = buildURL(path);
      if (!_disabled) {
        return navigate(url, {
          scroll,
          keepfocus,
          replace,
          state
        });
      }
      await goLocation(url);
    },
    setScrollDelegate(manager) {
      return _scrollDelegate = isFunction(manager) ? manager?.(router) : manager;
    },
    setRouteComparator(factory) {
      _comparator = factory() ?? createSimpleRoutesComparator();
    }
  };
  _scrollDelegate = createSimpleScrollDelegate(router);
  initialRoutes?.forEach((route) => {
    router.addRoute(route);
  });
  return router;
}
function getBaseUri(baseUrl = "/") {
  return import.meta.env.SSR ? `http://ssr${baseUrl}` : `${location.protocol}//${location.host}${baseUrl === "/" ? "" : baseUrl}`;
}
export {
  createRouter
};
