/// <reference types="urlpattern-polyfill" />

export declare type AppConfig = {
    id: string;
    module: {
        [id: string]: unknown;
    };
    baseUrl: string;
    configs: ConfigureApp[];
};

export declare type CancelNavigation = () => void;

export declare type ClientLayout = {
    /** Layout name. */
    readonly name: string;
    /** System file path relative to `<root>`. */
    readonly rootPath: string;
    /** Layout module loader. Used to dynamically import client-side. */
    readonly loader: () => Promise<ClientLayoutModule>;
};

export declare type ClientLayoutModule = ClientPageModule;

export declare type ClientLoadedData = Record<string, unknown>;

export declare type ClientPage = {
    /** System file path relative to `<root>`. */
    readonly rootPath: string;
    /** Page route object. */
    readonly route: Route;
    /** Page file extension.  */
    readonly ext: string;
    /** Additional page metadata. */
    readonly context?: Record<string, unknown>;
    /** Page layout name. */
    readonly layoutName?: string;
    /** Page layouts identifiers. */
    readonly layouts: number[];
    /** Page module loader. Used to dynamically import page module client-side. */
    readonly loader: () => Promise<ClientPageModule>;
};

export declare type ClientPageModule = {
    readonly [id: string]: unknown;
    readonly default: unknown;
    readonly meta?: MarkdownMeta;
    readonly loader?: ServerLoader;
};

export declare type ComplexRouteMatcher = (route: string, info: {
    filePath: string;
}) => string | null | undefined | void;

export declare type ComplexScrollDelegate = Required<ScrollDelegate> & {
    /**
     * Base scroll settings that are applied to all scrolls.
     */
    setBase(calc: () => ScrollBase): void;
    /**
     * Indicates how the browser should scroll when navigating to a new page.
     *
     * @defaultValue `'auto'`
     */
    setBehavior(hook: ScrollBehaviorHook): void;
};

export declare type ConfigureApp = (app: {
    router: Router;
}) => void | Promise<void>;

/**
 * The complex scroll delegate saves the scroll position of each page to ensure it's put back
 * when the user either navigates back, reloads, or re-opens the tab. In addition, it enables
 * you to perform complex scrolling logic during navigation whilst adjusting base position and
 * behavior.
 */
export declare const createComplexScrollDelegate: ScrollDelegateFactory<ComplexScrollDelegate>;

export declare function createMemoryHistory(): MemoryHistory;

export declare function createRouter({ history, baseUrl, trailingSlash, $navigation, $route, initialRoutes, serverContext, }: RouterOptions): Router;

/**
 * Default routes comparator. It doesn't sort or score routes and relies on the work being done
 * server-side. This comparator's main job is match routes to the given URL.
 */
export declare function createSimpleRoutesComparator(): RoutesComparator;

/**
 * The simple scroll delegate saves the scroll position of each page to ensure it's put back
 * when the user either navigates back, reloads, or re-opens the tab.
 */
export declare const createSimpleScrollDelegate: ScrollDelegateFactory;

/**
 * Uses the same route matching, scoring/ranking, and sorting algorithm as the server-side
 * implementation. This can be used when you're programtically creating routes client-side. Be
 * careful because this comparator weighs a ~few KB, hence why it's not the default.
 */
export declare function createURLPatternRouteMatcher(): RoutesComparator;

export declare type GoToRouteOptions = {
    scroll?: ScrollToTarget | null;
    keepfocus?: boolean;
    replace?: boolean;
    state?: any;
};

export declare function initRouter({ $route, $navigation, $pages, $layouts, $currentPage, serverContext, }: RouterInitOptions): Promise<Router>;

export declare function installURLPattern(): Promise<void>;

export declare function isLoadedMarkdownPage(page: unknown): page is LoadedClientMarkdownPage;

export declare function isLoadedPage(page: unknown): page is LoadedClientPage;

export declare function loadData(router: Router, module: ClientPageModule | ClientLayoutModule, layoutIndex?: number, prefetchURL?: URL): Promise<ClientLoadedData>;

export declare type LoadedClientLayout = ClientLayout & {
    readonly $$loaded: true;
    readonly module: ClientLayoutModule;
    readonly default: unknown;
    readonly data: ClientLoadedData;
};

export declare type LoadedClientMarkdownPage = LoadedClientPage & {
    readonly meta: MarkdownMeta;
};

export declare type LoadedClientPage = Omit<ClientPage, 'layouts'> & {
    readonly $$loaded: true;
    readonly module: ClientPageModule;
    readonly default: unknown;
    readonly layouts: LoadedClientLayout[];
    readonly data: ClientLoadedData;
};

export declare type LoadedRoute = MatchedRoute & {
    page: LoadedClientPage;
};

export declare function loadPage(router: Router, clientPage: ClientPage, clientLayouts: ClientLayout[], $currentPage: Reactive<LoadedClientPage>, { prefetch }?: {
    prefetch?: boolean | URL;
}): Promise<string | LoadedClientPage>;

export declare function loadPageLayouts(router: Router, page: ClientPage, layouts: ClientLayout[], prefetchURL?: URL): Promise<LoadedClientLayout[]>;

export declare type MarkdownFrontmatter = Record<string, any>;

export declare type MarkdownHeading = {
    level: number;
    title: string;
    id: string;
};

export declare type MarkdownMeta = {
    title?: string | null;
    headings: MarkdownHeading[];
    frontmatter: MarkdownFrontmatter;
    lastUpdated: number;
};

export declare type MatchedRoute = RouteDeclaration & {
    readonly id: string;
    readonly url: URL;
    readonly params: RouteParams;
};

export declare type MaybeServerLoadedOutput<Data = ServerLoadedData> = void | undefined | null | ServerLoadedOutput<Data>;

/** Used server-side.  */
export declare class MemoryHistory implements History {
    scrollRestoration: ScrollRestoration;
    protected history: string[];
    protected position: number;
    protected _state: {};
    get state(): {};
    get length(): number;
    back(): void;
    forward(): void;
    go(delta: number): void;
    pushState(data: any, _: string, url?: string | URL | null): void;
    replaceState(state: any, _: string, url?: string | URL | null): void;
    protected setLocation(location: string): void;
}

export declare type NavigationOptions = GoToRouteOptions & {
    accepted?: () => void;
    blocked?: () => void;
};

export declare type Reactive<T> = {
    get(): T;
    set(newValue: T): void;
    subscribe(onUpdate: (value: T) => void): Unsubscribe;
};

export declare type ReactiveFactory = {
    <T>(value: T): Reactive<T>;
};

export declare type Route = {
    /** Order number if declared (e.g., `[1]page.md` would be 1). */
    readonly order?: number;
    /**
     * A positive integer representing the path match ranking. The route with the highest score
     * will win if the path matches multiple routes.
     */
    readonly score: number;
    /**
     * `URLPattern` used to match a pattern against a route.
     *
     * @see {@link https://developer.mozilla.org/en-US/docs/Web/API/URL_Pattern_API}
     */
    readonly pattern: URLPattern;
    /**
     * The pathname used to construct the `URLPattern`.
     */
    readonly pathname: string;
    /**
     * Whether the route pattern is dynamic. This includes wildcards `*`,
     * named groups `/:id`, non-capturing groups `{/path}` and RegExp groups `(\\d+)`.
     *
     * @see {@link https://developer.mozilla.org/en-US/docs/Web/API/URL_Pattern_API#pattern_syntax}
     */
    readonly dynamic: boolean;
};

export declare type RouteDeclaration = Route & {
    prefetch?: RoutePrefetch;
    loader: RouteLoader;
};

export declare type RouteLoader = (info: {
    route: MatchedRoute;
    redirect: RouteRedirect;
}) => void | LoadedClientPage | Promise<void | LoadedClientPage>;

export declare type RouteMatcher = string | RegExp | null | undefined | void;

export declare type RouteMatcherConfig = (SimpleRouterMatcher | ComplexRouteMatcher)[];

export declare type RouteNavigation = {
    from: URL;
    to: URL;
} | null;

export declare type RouteParams = Record<string, any>;

export declare type RoutePrefetch = (info: {
    url: URL;
    route: MatchedRoute;
    redirect: RouteRedirect;
}) => void | Promise<void>;

export declare type Router = {
    /**
     * The current URL.
     */
    readonly url: URL;
    /**
     * Manager used to handle scrolling-related tasks. A manager doesn't exist by default and needs
     * to be added via `addScrollManager(...)`
     */
    readonly scroll: ScrollDelegate;
    /**
     * Whether the router is disabled. Disabling the router means the browser will handle all
     * navigation and calling `goto` will be a no-op.
     *
     * @defaultValue `false`
     */
    disabled: boolean;
    /**
     * The base URL for all routes.
     *
     * @defaultValue `'/'`
     */
    readonly baseUrl: string;
    /**
     * Available during SSR.
     */
    readonly serverContext?: ServerContext;
    /**
     * Whether a slash should be appended to the end of HTML routes. This is modified by adapters
     * accordingly by injecting `__VBK_TRAILING_SLASH__` into the `window` object.
     *
     * @defaultValue `true`
     */
    readonly trailingSlash: boolean;
    /**
     * Reactive route navigation.
     */
    readonly navigation: Reactive<RouteNavigation>;
    /**
     * Whether the router is in the process of navigating to another page.
     */
    readonly navigating: boolean;
    /**
     * Whether the router has started (i.e., loaded first page).
     */
    readonly started: boolean;
    /**
     * The currently loaded route.
     */
    readonly currentRoute: LoadedRoute | null;
    /**
     * Start the router and begin listening for link clicks we can intercept them and handle SPA
     * navigation.
     */
    start(): void;
    /**
     * Called when navigating to a new route and right before a new page is loaded. Returning a
     * redirect path will navigate to the matching route declaration.
     *
     * @defaultValue undefined
     */
    beforeNavigate(hook: RouterBeforeNavigateHook): void;
    /**
     * Called after navigating to a new route and it's respective page has loaded.
     *
     * @defaultValue undefined
     */
    afterNavigate(hook: RouterAfterNavigateHook): void;
    /**
     * Redirect from a given pathname to another.
     */
    addRedirect(from: string | URL, to: string | URL): void;
    /**
     * Returns a route declaration given a URL pathname such as `/` or `/getting-started/intro.html`.
     */
    getRoute(pathname: string): WithRouteParams<RouteDeclaration> | null;
    /**
     * Returns whether the given pathname matches any route.
     */
    hasRoute(pathnameOrRoute: string | UserRouteDeclaration): boolean;
    /**
     * Registers a new route given a declaration.
     */
    addRoute(route: Omit<UserRouteDeclaration, 'score'> & {
        score?: number;
    }): void;
    /**
     * Deregisters a route given it's declaration.
     */
    removeRoute(route: RouteDeclaration): void;
    /**
     * Builds and returns a normalized application URL.
     */
    buildURL(pathnameOrURL: string | URL): URL;
    /**
     * Checks whether the given URL belongs to this application.
     */
    owns(url: URL): boolean;
    /**
     * Attempts to find, build, and return a `Route` object given a pathname or URL.
     */
    findRoute(pathnameOrURL: string | URL): MatchedRoute | null;
    /**
     * Navigate to the previous page.
     */
    back(): void;
    /**
     * Attempts to match the given path to a declared route and navigate to it. The path can be a
     * URL pathname (e.g., `/a/path.html`), hash (e.g., `#some-id`), or URL instance
     * (e.g., `new URL(...)`).
     */
    go(path: string | URL, options?: GoToRouteOptions): Promise<void>;
    /**
     * Loads `href` the old-fashioned way, with a full page reload. Returns a `Promise` that never
     * resolves to prevent any subsequent work (e.g., history manipulation).
     */
    goLocation(url: URL): Promise<void>;
    /**
     * Attempts to find a route given a pathname or URL and call it's prefetch handler. This method
     * will throw if no route matches the given pathname.
     */
    prefetch(pathnameOrURL: string | URL): Promise<void>;
    /**
     * Set a new target for delegating scroll tasks to.
     */
    setScrollDelegate<T extends ScrollDelegate>(manager: T | ScrollDelegateFactory<T>): T;
    /**
     * Adds a new route comparator to handle matching, scoring, and sorting routes.
     */
    setRouteComparator(factory: RoutesComparatorFactory): void;
    /**
     * Notifies the router of a hash change.
     */
    hashChanged(hash: string): any;
    /**
     * Key used to save navigation state in history state object.
     */
    readonly _historyKey: string;
    /**
     * Keeps track of the history index in order to prevent popstate navigation
     * events (if needed).
     */
    _historyIndex: number;
    /**
     * Internal navigation method.
     */
    _navigate(url: URL, options?: NavigationOptions): Promise<void>;
};

export declare type RouterAfterNavigateHook = (navigation: {
    from: LoadedRoute | null;
    to: LoadedRoute;
    params: RouteParams;
}) => void | Promise<void>;

export declare type RouterBeforeNavigateHook = (navigation: {
    from: LoadedRoute | null;
    to: WithRouteParams<RouteDeclaration>;
    params: RouteParams;
    cancel: CancelNavigation;
    redirect: RouteRedirect;
}) => void;

export declare type RouteRedirect = (pathnameOrURL: string | URL) => void;

export declare type RouterInitOptions = {
    $route: Reactive<LoadedRoute>;
    $navigation: Reactive<RouteNavigation>;
    $pages: Reactive<ClientPage[]>;
    $layouts: Reactive<ClientLayout[]>;
    $currentPage: Reactive<LoadedClientPage>;
    serverContext?: ServerContext;
};

export declare type RouterOptions = {
    baseUrl: string;
    history: History;
    trailingSlash?: boolean;
    $route: Reactive<LoadedRoute>;
    $navigation: Reactive<RouteNavigation>;
    initialRoutes?: RouteDeclaration[];
    serverContext?: ServerContext;
};

export declare type RoutesComparator = {
    /**
     * Matches the URL to a route in the given `routes` list.
     */
    match(url: URL, routes: RouteDeclaration[]): WithRouteParams<RouteDeclaration> | null;
    /**
     * Returns a score for ranking the given route. Routes with a higher score should be matched
     * before routes with a lower score.
     */
    score(route: UserRouteDeclaration): number;
    /**
     * Sorts the routes list into the order in which they should be matched. Routes at earlier
     * positions should match first.
     */
    sort(routes: ScoredRouteDeclaration[]): ScoredRouteDeclaration[];
};

export declare type RoutesComparatorFactory = () => RoutesComparator;

export declare type ScoredRouteDeclaration = RouteDeclaration & {
    readonly score: number;
};

export declare type ScrollBase = ScrollToOptions;

export declare type ScrollBehaviorHook = (info: {
    from: LoadedRoute | null;
    to: LoadedRoute;
    cancel: ScrollCancel;
    savedPosition?: {
        top?: number;
        left?: number;
    };
}) => ScrollTarget | Promise<ScrollTarget>;

export declare type ScrollCancel = () => void;

export declare type ScrollDelegate = {
    /**
     * Scrolls to the given target element, hash, or position.
     */
    scroll?(options: ScrollOptions_2): void;
    /**
     * Called by the router at certain points such as before navigating to a new page or the
     * URL hash changing.
     */
    savePosition?(): void;
    /**
     * Expected to return the saved scroll position given the history state object.
     */
    getSavedPosition?(state: Record<string, any>): ScrollPosition;
    /**
     * Called when scroll positions should be persisted (i.e., local/session storage).
     */
    commit?(): void;
};

export declare type ScrollDelegateFactory<T extends ScrollDelegate = ScrollDelegate> = (router: Router) => T;

declare type ScrollOptions_2 = {
    hash?: string;
    from?: LoadedRoute | null;
    to?: LoadedRoute;
    target?: ScrollToTarget | null;
};
export { ScrollOptions_2 as ScrollOptions }

export declare type ScrollPosition = {
    top: number;
    left: number;
};

export declare function scrollPosition(): ScrollPosition;

export declare type ScrollTarget = void | null | false | (ScrollToOptions & {
    el?: string | HTMLElement;
});

export declare type ScrollToTarget = (info: {
    cancel: ScrollCancel;
}) => ScrollTarget;

export declare type ServerContext = {
    modules: Set<string>;
    data: ServerLoadedDataMap;
};

export declare type ServerEndpoint = ServerFile & {
    /** Routing object. */
    readonly route: Route;
};

export declare type ServerEntryModule = {
    render: ServerRenderer;
};

export declare type ServerFile = {
    /** Absolute system file path to file.  */
    readonly filePath: string;
    /** System file path relative to `<root>` to associated file. */
    readonly rootPath: string;
};

export declare type ServerLayout = ServerFile & Omit<ClientLayout, 'loader'> & {
    /** Module id used by the client-side router to dynamically load this layout module.  */
    id: string;
    /** The root directory that this layout belongs to. */
    readonly owningDir: string;
    /** Whether the layout has a data `loader` function. */
    hasLoader: boolean;
    /** Whether the current layout resets the layout stack.  */
    reset: boolean;
};

export declare type ServerLoadedData = Record<string, unknown>;

/** Map of data asset id to server loaded data object. */
export declare type ServerLoadedDataMap = Map<string, ServerLoadedData>;

export declare type ServerLoadedOutput<Data = ServerLoadedData> = {
    data?: Data;
    readonly redirect?: string | {
        path: string;
        statusCode?: number;
    };
    readonly cache?: ServerLoaderCacheKeyBuilder;
};

/** Map of data asset id to server loaded output object. */
export declare type ServerLoadedOutputMap = Map<string, ServerLoadedOutput>;

export declare type ServerLoadedRedirect = {
    path: string;
    statusCode: number;
};

export declare type ServerLoader<Data = ServerLoadedData> = (input: ServerLoaderInput) => MaybeServerLoadedOutput<Data> | Promise<MaybeServerLoadedOutput<Data>>;

/** Key can be anything but only truthy values are used to cache. */
export declare type ServerLoaderCacheKey = unknown;

export declare type ServerLoaderCacheKeyBuilder = (input: ServerLoaderInput) => ServerLoaderCacheKey | Promise<ServerLoaderCacheKey>;

export declare type ServerLoaderCacheMap = Map<ServerLoaderCacheKey, ServerLoadedOutput>;

export declare type ServerLoaderInput<Params extends ServerLoaderParams = ServerLoaderParams> = Readonly<{
    pathname: string;
    page: ServerPage;
    route: Route;
    params: Params;
    /** Result from running `URLPattern.exec().pathname`. */
    match: URLPatternComponentResult;
}>;

export declare type ServerLoaderParams = {
    [param: string]: string | undefined;
};

export declare type ServerPage = ServerFile & Omit<ClientPage, 'loader' | 'layouts'> & {
    /** Module id used by the client-side router to dynamically load this page module.  */
    id: string;
    /** Routing object. */
    readonly route: Route;
    /** Page layout name. */
    layoutName?: string;
    /**
     * Indentifies layout files that belong to this page. Each number is an index to a layout
     * client layout file in the `layouts` store.
     */
    layouts: number[];
    /**
     * Additional data to be included with the page. This will be included in the client-side
     * response.
     */
    context: Record<string, unknown>;
    /** Whether the page has a data `loader` function. */
    hasLoader: boolean;
};

export declare type ServerRenderer = (url: URL, context: {
    data: ServerContext['data'];
}) => Promise<ServerRenderResult>;

export declare type ServerRenderResult = {
    head?: string;
    css?: string;
    html: string;
    context: ServerContext;
};

export declare type SimpleRouterMatcher = {
    name: string;
    matcher: RouteMatcher;
};

export declare type Unsubscribe = () => void;

export declare type UserRouteDeclaration = Omit<RouteDeclaration, 'score'> & {
    score?: number;
};

export declare type WithRouteMatch<T> = T & {
    match: URLPatternComponentResult;
};

export declare type WithRouteParams<T, U = RouteParams> = T & {
    params: U;
};

export { }
